/*************************************************************
 *
 *  dla3d.c - Three-dimensional diffusion-limited aggregation
 *
 *  Copyright (C) 2000-2004  Mark J. Stock, mstock@umich.edu
 * 
 *  This file is part of dla3d.
 *
 *  dla3d is free software; you can redistribute it and/or modify
 *  it under the terms of the GNU General Public License as published by
 *  the Free Software Foundation; either version 2 of the License, or
 *  (at your option) any later version.
 *
 *  dla3d is distributed in the hope that it will be useful,
 *  but WITHOUT ANY WARRANTY; without even the implied warranty of
 *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 *  GNU General Public License for more details.
 *
 *  You should have received a copy of the GNU General Public License
 *  along with dla3d; if not, write to the Free Software
 *  Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
 * 
 * Revision information
 *
 * 2003-12-03 MJS  v0.0  copied part3d source, begin work
 * 2003-12-04 MJS  v0.1  improved order over O(N^2)
 * 2003-12-04 MJS  v0.2  new output, correct aggregation
 * 2004-01-20 MJS  v0.3  added directionality, reorganized
 *
 *********************************************************** */


#include "structs.h"

int run_sim(fileprop_ptr,sim_ptr,cell_ptr);
int diffuse_new_particle(sim_ptr,cell_ptr,FLOAT*,FLOAT*);
void pick_point_on_sphere(FLOAT*,FLOAT*,FLOAT,FLOAT*);
FLOAT find_dist_to_closest(cell_ptr,FLOAT*,FLOAT,particle_ptr);
FLOAT find_dist_to_farthest(cell_ptr,FLOAT*,FLOAT);
void sum_tipward_mass(cell_ptr);
void zero_all_masses(cell_ptr);


int main(int argc,char **argv) {

   FILEP file_props;		// initialize struct for file props
   fileprop_ptr file = &file_props;
   SIMP sim_props;		// initialize struct for simulation props
   sim_ptr sim = &sim_props;
   CELL top_cell;		// initialize struct for octree
   cell_ptr top = &top_cell;

   // set some important pointers
   file->out = &file->out_field;
   sim->ff2 = &sim->density_field2;
   sim->ff3 = &sim->density_field3;

   /* Parse command-line args */
   parse_args(argc,argv,file);

   /* Initialize the top-level cell and participating particles */
   initialize_system(file,sim,top);

   /* Do the time integration */
   run_sim(file,sim,top);

   fprintf(stderr,"\nDone.\n");
   exit(0);
}



/*
 *  Run the time integration of the flowfield simulation
 */
int run_sim(fileprop_ptr file,sim_ptr sim,cell_ptr top) {

   int i;
   FLOAT start[3],vel[3];

   sim->next_output_step = sim->output_step;

   // loop over max number of particles
   for (sim->step=1;sim->step<sim->particle_cnt+1;sim->step++) {

      // Write some stuff
      if (sim->step%1000 == 0) {
         fprintf(stdout,"\nStep %d, next print %g, %d particles within %g radius\n",sim->step,sim->next_output_step,top->num,sim->overall_rad);
      }
      // write particle count and cell utilization
      if (sim->step%10000 == 0) {
         fprintf(stdout,"  %d particles\n",top->num);
         for (i=0; i<sim->max_levels; i++) sim->cell_count[i] = 0;
         count_cells(sim,top);
         fprintf(stdout,"  ");
         for (i=0; i<sim->max_levels; i++) fprintf(stdout,"%d ",sim->cell_count[i]);
         fprintf(stdout,"cells\n  utilization ");
         for (i=0; i<sim->max_levels; i++) fprintf(stdout,"%g ",sim->cell_count[i]/pow(8,i));
         fprintf(stdout,"\n");
      }
      if (sim->step%100 == 0) {
         fprintf(stdout,"."); fflush(stdout);
      }

      // check current step vs. next_output_step, write output if necessary
      if (sim->step == sim->next_output_step) {
         fprintf(stdout,"\n  writing output step %d\n",sim->next_output_index);
         write_output(file,sim,top);
         sim->next_output_step += sim->output_step;
         sim->next_output_index++;
      }

      // place a new particle
      // diffuse_new_particle(sim,top,0.0,0.0,0.0);
      start[0]=0.0;
      start[1]=0.0;
      start[2]=0.0;
      vel[0]=0.0;
      vel[1]=0.0;
      vel[2]=-1000.0;
      diffuse_new_particle(sim,top,start,vel);

      /* remove any cells with no particles, and combine cells with
       * too few particles, it also recalculates the proper number of
       * particles per cell */
      clean_up_all_cells(sim,top);
   }

   // return code 1 - ran out of particles
   fprintf(stdout,"Simulation completed.\n");
   return(1);
}


/*
 *  Run the random-walk method until a hit occurs
 */
int diffuse_new_particle(sim_ptr sim,cell_ptr top,FLOAT* start,FLOAT* vel) {

   int i,j,trying_particles,still_walking;
   FLOAT loc[3],center[3],rad,initvel[3];
   particle_ptr newpart;

   // use negative velocity for initial particle placement
   // for (j=0;j<3;j++) initvel[j]=-1.0*vel[j];
   for (j=0;j<3;j++) initvel[j]=0.0;

   // create the new part right here
   newpart = new_stationary_particle(sim->step,sim->new_part_rad,start[0],start[1],start[2]);

   trying_particles = TRUE;
   while (trying_particles) {

      // find particle start point
      // rad = SQRT2*(top->max[0]-top->min[0])/2.;
      // rad = find_bounding_sphere(top);
      // loc[0] = top->mid[0];
      // loc[1] = top->mid[1];
      // loc[2] = top->mid[2];
      loc[0] = start[0];
      loc[1] = start[1];
      loc[2] = start[2];

      // take one step to start
      rad = find_dist_to_farthest(top,loc,0.0)+3.6*sim->new_part_rad;
      // rad = find_dist_to_closest(top,loc,LARGE,newpart)-2.0*sim->new_part_rad;
      sim->overall_rad = rad;
      // fprintf(stderr,"  maxrad is %g\n",rad);
      // fprintf(stderr,"starting at %g %g %g\n",loc[0],loc[1],loc[2]);
      for (j=0;j<3;j++) center[j]=loc[j];
      pick_point_on_sphere(loc,center,rad-1.8*sim->new_part_rad,initvel);
      rad = find_dist_to_closest(top,loc,LARGE,newpart);
      if (rad < 2.0*sim->new_part_rad)
         still_walking = FALSE;
      else
         still_walking = TRUE;

      // loop until particle hits a static particle
      while (still_walking) {

         // move the particle to a sphere of that radius
         for (j=0;j<3;j++) center[j]=loc[j];
         pick_point_on_sphere(loc,center,rad-1.8*sim->new_part_rad,vel);
         // fprintf(stderr,"  point at %g %g %g\n",loc[0],loc[1],loc[2]);

         // check for contact between this part and a static one
         rad = find_dist_to_closest(top,loc,LARGE,newpart);
         // fprintf(stderr,"  nearest is %g\n\n",rad);
         if (rad < 2.0*sim->new_part_rad || rad > 100.*sim->overall_rad)
            still_walking = FALSE;
      }

      // if we quit because we were too far away...
      if (rad < 2.0*sim->new_part_rad) {
         // only now locate the particle
         newpart->x[0] = loc[0];
         newpart->x[1] = loc[1];
         newpart->x[2] = loc[2];
         add_particle_to_cell(sim,newpart,top);
         // fprintf(stderr,"  planting particle at %g %g %g, rad is %g, npr is %g\n\n",loc[0],loc[1],loc[2],rad,sim->new_part_rad);

         trying_particles = FALSE;
      } else {
         // fprintf(stderr,"  particle at %g %g %g, rad is %g, is too far\n",loc[0],loc[1],loc[2],rad);
      }
   }

   return(1);
}


/*
 * find a random point on a given sphere shell
 */
void pick_point_on_sphere(FLOAT* loc,FLOAT* center,FLOAT rad,FLOAT* vel) {

   FLOAT len = 2.;
   FLOAT dt = rad*rad;

   while (len > 1.) {
      loc[0] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      loc[1] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      loc[2] = 2.*rand()/(RAND_MAX+1.0) - 1.;
      len = sqrt(loc[0]*loc[0]+loc[1]*loc[1]+loc[2]*loc[2]);
   }
   loc[0] += vel[0]*dt;
   loc[1] += vel[1]*dt;
   loc[2] += vel[2]*dt;
   len = sqrt(loc[0]*loc[0]+loc[1]*loc[1]+loc[2]*loc[2]);
   loc[0] = center[0]+rad*loc[0]/len;
   loc[1] = center[1]+rad*loc[1]/len;
   loc[2] = center[2]+rad*loc[2]/len;

   return;
}


/*
 *  Find the distance to the closest particle
 *  based on is_this_space_open2
 *
 *  cell is cell to search within
 *  loc is point to search against
 *  nearest is current value for nearest
 */
FLOAT find_dist_to_closest(cell_ptr cell,FLOAT loc[3],FLOAT neardist,
                           particle_ptr newpart){

   int i,j,k,sx,sy,sz;
   FLOAT dist,nsq;
   particle_ptr curr;

   // if (cell->level == 0) neardist = 9.99e+99;
   // neardist = 9.99e+99;
   nsq = neardist*neardist;

   if (loc[0] < cell->min[0]-neardist) return(neardist);
   if (loc[0] > cell->max[0]+neardist) return(neardist);
   if (loc[1] < cell->min[1]-neardist) return(neardist);
   if (loc[1] > cell->max[1]+neardist) return(neardist);
   if (loc[2] < cell->min[2]-neardist) return(neardist);
   if (loc[2] > cell->max[2]+neardist) return(neardist);

   if (cell->has_subcells) {
      // must check all of the subcells, but check closest first
      if (loc[0] < cell->mid[0]) sx=0; else sx=1;
      if (loc[1] < cell->mid[1]) sy=0; else sy=1;
      if (loc[2] < cell->mid[2]) sz=0; else sz=1;
      for (i=sx;i<sx+2;i++)
         for (j=sy;j<sy+2;j++)
            for (k=sz;k<sz+2;k++) {
               dist = find_dist_to_closest(cell->s[i%2][j%2][k%2],
                         loc,neardist,newpart);
               if (dist < neardist) neardist = dist;
               // fprintf(stderr,"    closest is %g\n",neardist);
            }
   } else {
      // check vs. all of the particles in this cell
      // fprintf(stderr,"    checking vs %d particles in box at level %d\n",cell->num,cell->level);
      curr = cell->first;
      while (curr) {
         // dist = sqrt(pow(loc[0]-curr->x[0],2)+
         //             pow(loc[1]-curr->x[1],2)+
         //             pow(loc[2]-curr->x[2],2));
         dist = pow(loc[0]-curr->x[0],2)+
                pow(loc[1]-curr->x[1],2)+
                pow(loc[2]-curr->x[2],2);
         // fprintf(stderr,"    dist to this particle is %g\n",dist);
         // if (dist < neardist) neardist = dist;
         if (dist < nsq) {
            nsq = dist;
            neardist = sqrt(dist);
            newpart->root = curr;
         }
         curr = curr->next;
      }
   }

   return(neardist);
}


/*
 *  Find the distance to the farthest particle
 *  based on is_this_space_open2
 */
FLOAT find_dist_to_farthest(cell_ptr cell,FLOAT loc[3],FLOAT farthest){

   int i,j,k;
   FLOAT dsq,fsq,dmin,dmax,dist;
   particle_ptr curr;

   // if (cell->level == 0) farthest = 0.;
   // farthest = 0.;
   fsq = farthest*farthest;

   // must see if whole cell is inside of sphere radius "farthest"
   // so, find the distance to the farthest corner on the cube
   dsq = 0.;
   for (i=0;i<3;i++) {
      dmin = fabs(loc[i]-cell->min[i]);
      dmax = fabs(loc[i]-cell->max[i]);
      if (dmin > dmax) dsq += dmin*dmin;
      else dsq += dmax*dmax;
   }
   if (dsq < fsq) return(farthest);

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++) {
               dist = find_dist_to_farthest(cell->s[i][j][k],loc,farthest);
               if (dist > farthest) farthest = dist;
               // fprintf(stderr,"    farthest is %g\n",farthest);
            }
   } else {
      // check vs. all of the particles in this cell
      // fprintf(stderr,"    checking vs %d particles in box at level %d\n",cell->num,cell->level);
      curr = cell->first;
      while (curr) {
         // dist = sqrt(pow(loc[0]-curr->x[0],2)+
         //             pow(loc[1]-curr->x[1],2)+
         //             pow(loc[2]-curr->x[2],2));
         dist = pow(loc[0]-curr->x[0],2)+
                pow(loc[1]-curr->x[1],2)+
                pow(loc[2]-curr->x[2],2);
         // fprintf(stderr,"    dist to this particle is %g\n",dist);
         // if (dist > farthest) farthest = dist;
         if (dist > fsq) {
            fsq = dist;
            farthest = sqrt(dist);
         }
         curr = curr->next;
      }
   }

   return(farthest);
}


/*
 * Starting with each particle, march toward the center
 * and add 1.0=mass to every node along the way
 */
void sum_tipward_mass(cell_ptr cell) {

   int i,j,k;
   particle_ptr curr,march;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++)
               sum_tipward_mass(cell->s[i][j][k]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->mass += 1.0;
         march = curr->root;
         while (march) {
            march->mass += 1.0;
            march = march->root;
         }
         curr = curr->next;
      }
   }

   return;
}


/*
 * zero all masses
 */
void zero_all_masses(cell_ptr cell) {

   int i,j,k;
   particle_ptr curr;

   if (cell->has_subcells) {
      // must check all of the subcells
      for (i=0;i<2;i++)
         for (j=0;j<2;j++)
            for (k=0;k<2;k++)
               zero_all_masses(cell->s[i][j][k]);
   } else {
      // check vs. all of the particles in this cell
      curr = cell->first;
      while (curr) {
         curr->mass = 0.0;
         curr = curr->next;
      }
   }

   return;
}
